package net.sf.jcc.state;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.omg.uml.Behavior;
import org.omg.uml.Constraint;
import org.omg.uml.NamedElement;
import org.omg.uml.OpaqueBehavior;
import org.omg.uml.OpaqueExpression;
import org.omg.uml.Pseudostate;
import org.omg.uml.PseudostateKind;
import org.omg.uml.Region;
import org.omg.uml.State;
import org.omg.uml.StateMachine;
import org.omg.uml.Transition;
import org.omg.uml.Trigger;

import net.sf.jcc.model.parser.uml2.ElementStore;
import net.sf.jcc.model.parser.uml2.ModelElement;


public class StateEngineImpl<Entity> implements StateEngine<Entity> 
{
	private static final String INITIAL_STATE = ".!intial";
	private ElementStore store;
	private String stateMachineIdentifier;
	private GuardMatch<Entity> guardMatch;
	private EffectEngine<Entity> effectEngine;
	
	private transient Map<String, State> statesByName;
	
	@Override
	public String initialise(Entity target, String trigger) 
	{
		return signal(target, INITIAL_STATE, trigger);
	}

	@Override
	public String signal(Entity target, String stateName, String trigger) 
	{
		State state = getState(stateName);
		
		List<Transition> transitions = resolveTransitions(state);
		
		for (Transition transition : transitions)
		{
			if (matchTransition(transition, target, trigger))
			{
				raiseEffect(transition, target);
				String targetStateRef = transition.getTarget();
				ModelElement element = store.get(targetStateRef);
				return ((State)element.getElement()).getName();
			}
		}
		return stateName;
	}

	protected void raiseEffect(Transition transition, Entity target) 
	{
		if (effectEngine == null)
		{
			return;
		}
		for (Behavior behavior : transition.getEffect())
		{
			List<String> bodies = ((OpaqueBehavior)behavior).getBody();
			if (bodies.isEmpty())
			{
				throw new RuntimeException("ModelValidation: Transition:" + transition.getName() + " does not have an expression.");
			}
			if (bodies.size() > 1)
			{
				throw new RuntimeException("ModelValidation: Transition:" + transition.getName() + " has more than one expression.");
			}
			effectEngine.invokeEffect(target, bodies.get(0));
		}
	}

	public void setStore(ElementStore store) 
	{
		this.store = store;
	}

	public void setGuardMatch(GuardMatch<Entity> guardMatch) 
	{
		this.guardMatch = guardMatch;
	}

	public void setStateMachineIdentififer(String stateMachineIdentifier) 
	{
		this.stateMachineIdentifier = stateMachineIdentifier; 
	}

	public void setEffectEngine(EffectEngine<Entity> effectEngine) 
	{
		this.effectEngine = effectEngine;
	}
	
	protected State getState(String stateName)
	{
		synchronized(this)
		{
			if (statesByName == null)
			{
				loadDefinition();
			}
		}
		
		return statesByName.get(stateName);
	}
	
	protected void loadDefinition()
	{
    	ModelElement stateMachineElement = store.get(stateMachineIdentifier);
    	if (stateMachineElement == null)
    	{
    		throw new RuntimeException("ModelValidation: GUID not found:" + stateMachineIdentifier);
    	}
    	StateMachine stateMachine = (StateMachine) stateMachineElement.getElement();
    	Region region = stateMachine.getRegion();
    	statesByName = new HashMap<String, State>();
    	for (State state : region.getSubvertex())
    	{
    		if (state instanceof Pseudostate)
    		{
    			if (((Pseudostate)state).getKind() == PseudostateKind.INITIAL)
    			{
    	    		statesByName.put(INITIAL_STATE, state);
    				continue;
    			}
    		}
    		statesByName.put(state.getName(), state);
    	}
	}
	
	private List<Transition> resolveTransitions(State state) 
	{
		synchronized(state)
		{
			List<Transition> transitions = state.getOutgoing();
			if (transitions.size() != 0)
			{
				return transitions;
			}
			
			
	    	ModelElement stateElement = store.reverse(state);

	    	List<String> refs = stateElement.getReferences(new QName("outgoing"));

            if (refs == null)
            {
                throw new RuntimeException("ModelValidation: State:" + state.getName() + " does not have an outgoing transition.");
            }
	    	
	    	for (String ref : refs)
	    	{
		    	ModelElement outgoing = store.get(ref);
		    	Transition transition = (Transition) outgoing.getElement();
		    	transitions.add(transition);
	    	}
			return transitions;
		}
	}

	private List<Trigger> resolveTriggers(Transition transition) 
	{
		synchronized(transition)
		{
			List<Trigger> triggers = transition.getTrigger();
			if (triggers.size() != 0)
			{
				return triggers;
			}
			
			
	    	ModelElement transitionElement = store.reverse(transition);

	    	List<String> refs = transitionElement.getReferences(new QName("trigger"));
	    	
	    	if (refs == null)
	    	{
	    		return triggers;
	    	}
	    	
	    	for (String ref : refs)
	    	{
		    	ModelElement outgoing = store.get(ref);
		    	Trigger trigger = (Trigger) outgoing.getElement();
		    	triggers.add(trigger);
	    	}
			return triggers;
		}
	}
	
	protected <T extends NamedElement> T findElement(List<T> elements, String name)
	{
		for (NamedElement element : elements)
		{
			if (element.getName().equals(name))
			{
				return (T)element;
			}
		}
		return null;
	}
	
	protected boolean matchTransition(Transition transition, Entity target, String triggerName) 
	{
		if (triggerName != null)
		{
			List<Trigger> triggers = resolveTriggers(transition);
			Trigger trigger = findElement(triggers, triggerName);
			if (trigger == null)
			{
				// This transition does not support this trigger
				return false;
			}
		}
		List<Constraint> guards = transition.getGuard();
		if (guards.isEmpty())
		{
			return true;
		}
		for (Constraint guard : guards)
		{
			List<String> bodies = ((OpaqueExpression)guard.getSpecification()).getBody();
			if (bodies.isEmpty())
			{
				throw new RuntimeException("ModelValidation: Guard:" + guard.getName() + " does not have an expression.");
			}
			if (bodies.size() > 1)
			{
				throw new RuntimeException("ModelValidation: Guard:" + guard.getName() + " has more than one expression.");
			}
			if (guardMatch.match(target, bodies.get(0)))
			{
				return true;
			}
		}
	
		return false;
	}
}
